function output = SingleMigration_EIDOptimal_3Param(numberOfPatches, distanceMatrix, samplingPower, powerDecayRate, totalPower, mu, delta, theta, maxVarianceExit, maxTimeSteps, fixIC, numberOfIndTimes, threeOrTwo)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Pagendam and Ross, 2011, Optimal GPS tracking for estimating species
%movements
%CODE for finding an EID-optimal design for the 3 parameter metapopulation model
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%EXAMPLE for entering data
%numberOfPatches = 4; The number of habitat patches in the landscape.
%distanceMatrix = [0 3 4 5; 3 0 1 6; 4 1 0 4; 5 6 4 0]; The pairwise distances between the habitat patches.
%samplingPower = (1.1925E-4)*120; This is the amount of power that is used when taking a sample
%powerDecayRate = 3.0763E-8*(24*3600); This is the exponential decay rate of the battery power with time.
%totalPower = 1.0; This is the total power of the battery
%mu = lognrnd(0.1, 0.2, 1, 50); %A vector of draws from the prior distribution over mu. Only 50 values are used here and we have used a lognormal prior.
%delta = lognrnd(0.05, 0.01, 1, 50); %A vector of draws from the prior distribution over delta. Only 50 values are used here and we have used a lognormal prior.
%theta = lognrnd(-0.7, 0.5, 1, 50); %A vector of draws from the prior distribution overtheta. Only 50 values are used here and we have used a lognormal prior.
%maxVarianceExit = 0.01; Terminate the cross-entropy maximisation if the maximum variance of the distributions is less than this.
%maxTimeSteps = 20; This is the maximum number of cross-entropy iterations to run
%fixIC = 0; If equal to 0, the optimal initial patch will be determined.  Otherwise, change this to the integer of the patch you want the individual to start in at time 0.
%numberOfIndTimes = 5;  This value ditates how many of the initial observation times will be drawn independently.  All others will be equispaced.  
%So in this example, the first 5 observation times will be independently arranged, but all subsequent observations will occur equispaced in time.
%threeOrTwo = 1; If set to 1, find the optimal design for estimating all three parameters.  If set to 0, find the optimal design for estimating mu and delta only.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


tdp = 0.001;
numberOfTimePoints = floor(totalPower/samplingPower);
numberOfVectors = 100;
alphaQuantile = 0.85;
smoothUpdater = 0.7;
dynamicSmoothUpdater = 0.7;
dynamicSmoothExponent = 5;
timeStep = 0;
p_fudge = 1/numberOfPatches/10;
tMax = (log(totalPower) - log(samplingPower))/powerDecayRate;
len_delta = length(delta);

%initialize the beta densities to be uniform on the time interval
%alphas = ones(numberOfIndTimes+1,1);
%betas = ones(numberOfIndTimes+1,1);

schedulesDiffs = betarnd(1,1,numberOfVectors,numberOfIndTimes+1);
schedulesDiffstMax = schedulesDiffs*tMax/numberOfTimePoints;
schedules = cumsum([schedulesDiffstMax, schedulesDiffstMax(:,end)*ones(1,numberOfTimePoints-numberOfIndTimes-1)],2);
starts = ceil(rand(1,numberOfVectors)*numberOfPatches);

%create an array to hold the performance of each schedule
performance = zeros(numberOfVectors,1);

%calculate the performance of each of our schedules
for(i = 1:numberOfVectors)
    i
    %kill off any observations that aren't within the battery power
    %constraints
    powerAtSample = zeros(1, numberOfTimePoints);
    startPower = totalPower;
    timeDiff = diff([0, schedules(i,:)]);
    for(j = 1:numberOfTimePoints)
        powerAtSample(j) = startPower*exp(-powerDecayRate*timeDiff(j)) - samplingPower;
        startPower = powerAtSample(j);
        %%%
        if powerAtSample(j) <= 0
            powerAtSample(j:numberOfTimePoints) = -1;
            break
        end
    end
    indices = find(powerAtSample > 0);
    if(fixIC > 0)
        for(jj=1:length(delta))
			if(length(indices) > 0 && sum(timeDiff>tdp*ones(1,length(timeDiff)))==length(timeDiff))
				performance(i) = performance(i) + (1/det(SingleMigrationFIM_3Param_Josh(distanceMatrix, mu(jj), delta(jj), theta(jj), fixIC, schedules(i,indices), threeOrTwo)))/len_delta;
			else
				performance(i) = Inf; %We do this because we are going to invert the performances below
			end
		end
    else
        for(jj=1:length(delta))
            jj
			if(length(indices) > 0 && sum(timeDiff>tdp*ones(1,length(timeDiff)))==length(timeDiff))
				performance(i) = performance(i) + (1/det(SingleMigrationFIM_3Param_Josh(distanceMatrix, mu(jj), delta(jj), theta(jj), starts(i), schedules(i,indices), threeOrTwo)))/len_delta;
			else
				performance(i) = Inf; %We do this because we are going to invert the performances below
			end
		end
    end
    if(performance(i) == 0)
        'ERROR!'
        pause
    end
    performance(i) = 1/performance(i);
end

%take the cream of the crop
performanceQuantile = quantile(performance, alphaQuantile);
indexVector = find(performance >= performanceQuantile);

%create the new values of alpha and beta from the cream of the crop
bestSchedulesDiffs = zeros(length(indexVector),numberOfIndTimes+1);
%%%
%for(i = 1:length(indexVector));
bestSchedulesDiffs(1:length(indexVector),:) = schedulesDiffs(indexVector(1:length(indexVector)),:);  %NOTE TO SELF THIS CAN BE IMPROVED
%end

bestStarts = zeros(length(indexVector),1);
%%%
%for(i = 1:length(indexVector))
bestStarts(1:length(indexVector)) = starts(indexVector(1:length(indexVector)));		%NOTE TO SELF THIS CAN BE IMPROVED
%end

%%%
%varVector = zeros(numberOfTimePoints,1);
%meanVector = zeros(numberOfTimePoints,1);

for(i = 1:numberOfPatches)
    Start_p(i) = length(find(bestStarts == i))/(length(bestStarts)) + p_fudge;
end
Start_p = cumsum(Start_p/sum(Start_p));

%bestSchedules = bestSchedules/tMax;
%%%
%for(i = 1:numberOfTimePoints)
    varVector = var(bestSchedulesDiffs)';
    meanVector = mean(bestSchedulesDiffs)';
    q = (1-meanVector)./meanVector;
    alphas = (q./((1+q).^2) - varVector)./(varVector + q.*varVector);
    betas = (alphas - meanVector.*alphas)./meanVector;
%end


%Now continue to update the parameters until the variances for the times are small
%%%
%for(i = 1:numberOfTimePoints)
newVariances = alphas.*betas./(((alphas + betas).^2).*(alphas + betas + 1));
%end

maximumVariance = max(newVariances);
minimumVariance = min(newVariances);

timeStep = timeStep+1;

while(maximumVariance > (maxVarianceExit^2) && minimumVariance > 0.0 && timeStep <= maxTimeSteps)
    tic
    timeStep
    starts = zeros(numberOfVectors, numberOfPatches);
    for(i = 1:numberOfVectors)
        thisRand = rand;
		indices = find(Start_p >= thisRand);
		starts(i) = indices(1);
    end
    %Generate a large number of sampling schedules using beta densities
    schedules = zeros(numberOfVectors, numberOfTimePoints);
	%successfulScheduleCounter = 1;
    %while(successfulScheduleCounter < numberOfVectors)
    %    test = 1;
	%	schedules(successfulScheduleCounter,:) = tMax*betarnd(alphas,betas);
    %    for(j = 2:numberOfTimePoints)
    %        if(schedules(successfulScheduleCounter,j) - schedules(successfulScheduleCounter,j-1) < 0)
    %            test = 0;
    %            break
    %        end
    %    end
    %    if test~=0
    %        successfulScheduleCounter = successfulScheduleCounter + 1
    %    end
    %end
    
    schedulesDiffs = betarnd(ones(numberOfVectors,1)*alphas',ones(numberOfVectors,1)*betas');
    schedulesDiffstMax = schedulesDiffs*tMax/numberOfTimePoints;
    schedules = cumsum([schedulesDiffstMax, schedulesDiffstMax(:,end)*ones(1,numberOfTimePoints-numberOfIndTimes-1)],2);
    %for anc = 1:numberOfVectors
        %schedulesDiffs(anc,:) = betarnd(alphas,betas);
        %schedulesDiffstMax(anc,:) = schedulesDiffs(anc,:)*tMax/((numberOfIndTimes+numberOfTimePoints)/2);
        %schedules(anc,:) = cumsum([schedulesDiffstMax(anc,:), schedulesDiffstMax(anc,end)*ones(1,numberOfTimePoints-numberOfIndTimes-1)],2);
        
        %for(j = 2:numberOfTimePoints)
        %    if(schedules(anc,j) - schedules(anc,j-1) < tdp)
        %        schedules(anc,j:end) = schedules(anc,j) + tdp + tdp/10;
        %    end
        %end
        
   % end
    
    %create an array to hold the performance of each schedule
    performance = zeros(numberOfVectors,1);

    %calculate the performance of each of our schedules
    for(i = 1:numberOfVectors)
        %kill off any observations that aren't within the bounds of cutoffTimes
        powerAtSample = zeros(1, numberOfTimePoints);
        startPower = totalPower;
        timeDiff = diff([0, schedules(i,:)]);
        for(j = 1:numberOfTimePoints)
            powerAtSample(j) = startPower*exp(-powerDecayRate*timeDiff(j)) - samplingPower;
            startPower = powerAtSample(j);
            %%%
            if powerAtSample(j) <= 0
                powerAtSample(j:numberOfTimePoints) = -1;
                break
            end
        end
        indices = find(powerAtSample > 0);
        if(fixIC > 0)
            for(jj=1:length(delta))
				if(length(indices) > 0 && sum(timeDiff>tdp*ones(1,length(timeDiff)))==length(timeDiff))
					performance(i) = performance(i) + (1/det(SingleMigrationFIM_3Param_Josh(distanceMatrix, mu(jj), delta(jj), theta(jj), fixIC, schedules(i,indices), threeOrTwo)))/len_delta;
				else
					performance(i) = Inf; %We do this because we are going to invert the performances below
				end
			end
        else
            for(jj=1:length(delta))
				if(length(indices) > 0 && sum(timeDiff>tdp*ones(1,length(timeDiff)))==length(timeDiff))
					performance(i) = performance(i) + (1/det(SingleMigrationFIM_3Param_Josh(distanceMatrix, mu(jj), delta(jj), theta(jj), starts(i), schedules(i,indices), threeOrTwo)))/len_delta;
				else
					performance(i) = Inf; %We do this because we are going to invert the performances below
				end
			end
        end
        if(performance(i) == 0)
            'ERROR!'
            pause
        end
        performance(i) = 1/performance(i);
    end
    
    %take the cream of the crop
    performanceQuantile = quantile(performance, alphaQuantile);
    indexVector = find(performance >= performanceQuantile);

    %create the new values of alpha and beta from the cream of the crop
    bestStarts = [];
	bestSchedulesDiffs = zeros(length(indexVector),numberOfIndTimes+1);

    
    %%%
    %for(i = 1:length(indexVector))
    bestStarts = starts(indexVector(1:length(indexVector)));	%NOTE TO SELF THIS CAN BE IMPROVED
    %bestSchedules(1:length(indexVector),:) = schedules(indexVector(1:length(indexVector)),:);
    bestSchedulesDiffs(1:length(indexVector),:) = schedulesDiffs(indexVector(1:length(indexVector)),:);
    %end

    %%%
    %newVarVector = zeros(numberOfTimePoints,1);
    %newMeanVector = zeros(numberOfTimePoints,1);
    

    %bestSchedules = bestSchedules/tMax;
    %%%
    thisDynamicSmoothUpdater = dynamicSmoothUpdater - dynamicSmoothUpdater*(1-1/timeStep)^dynamicSmoothExponent;

    %for(i = 1:numberOfTimePoints)       
     newVarVector = var(bestSchedulesDiffs)';
     newMeanVector = mean(bestSchedulesDiffs)';
     varVector = ((1-thisDynamicSmoothUpdater).*sqrt(varVector) + thisDynamicSmoothUpdater.*sqrt(newVarVector)).^2;
     meanVector = (1-smoothUpdater).*meanVector + smoothUpdater.*newMeanVector;
     q = (1-meanVector)./meanVector;    
     alphas = (q./((1+q).^2) - varVector)./(varVector + q.*varVector);
       % if(alphas(i) < 0.01)
       %     alphas(i) = 0.01;
       % end
     posind = find(alphas < 0.01);
     alphas(posind) = 0.01;
     betas = (alphas - meanVector.*alphas)./meanVector;
       % if(betas(i) < 0.001)
       %     betas(i) = 0.001;
       % end
     posind = find(betas < 0.001);
     betas(posind) = 0.001;
    %end
    
    for(i = 1:numberOfPatches)
        Start_p(i) = length(find(bestStarts == i))/(length(bestStarts)) + p_fudge;
    end
    Start_p = cumsum(Start_p/sum(Start_p));

    maximumVariance = max(varVector);
    minimumVariance = min(varVector);
    timeStep = timeStep+1;
    startPower = totalPower;
    
    %schedulesDiffstMax(anc,:) = (schedulesDiffs(anc,:)./max(schedulesDiffs(anc,:)))*tMax;
    %schedules(anc,:) = cumsum([schedulesDiffstMax(anc,:),
    %schedulesDiffstMax(anc,end)*ones(1,numberOfTimePoints-numberOfIndTimes-1)],2);
    
    meanVectorMod = meanVector*tMax/numberOfTimePoints;
    meanVectorMod = cumsum([meanVectorMod; meanVectorMod(end)*ones(numberOfTimePoints-numberOfIndTimes-1,1)]);
    
    timeDiff = diff([0, meanVectorMod']);
    for(j = 1:numberOfTimePoints)
         powerAtSample(j) = startPower*exp(-powerDecayRate*timeDiff(j)) - samplingPower;
         startPower = powerAtSample(j);
         %%%
         if powerAtSample(j) <= 0
            powerAtSample(j:numberOfTimePoints) = -1;
            break
         end
    end
    indices = find(powerAtSample > 0);
    if(fixIC > 0)
        [meanVectorMod(indices)', fixIC]
    else
        [meanVectorMod(indices)', mode(bestStarts)]
%        det(SingleMigrationFIM_3Param_Josh(distanceMatrix, mu, delta, theta, mode(bestStarts), meanVectorMod(indices)', threeOrTwo))
    end
 toc   
end

startPower = totalPower;
meanVectorMod = meanVector*tMax/numberOfTimePoints;
meanVectorMod = cumsum([meanVectorMod; meanVectorMod(end)*ones(numberOfTimePoints-numberOfIndTimes-1,1)]);
timeDiff = diff([0, meanVectorMod']);
for(j = 1:numberOfTimePoints)
    powerAtSample(j) = startPower*exp(-powerDecayRate*timeDiff(j)) - samplingPower;
    startPower = powerAtSample(j);
    if powerAtSample(j) <= 0
            powerAtSample(j:numberOfTimePoints) = -1;
            break
    end
end
indices = find(powerAtSample > 0);
if(fixIC > 0)
     output = [meanVectorMod(indices)', fixIC];
else
     output = [meanVectorMod(indices)', mode(bestStarts), timeStep];
end
